from django.test import SimpleTestCase

from django_perf_rec import pytest_plugin
from tests.utils import pretend_not_under_pytest


class PytestPluginTests(SimpleTestCase):
    def test_in_pytest(self):
        # We always run our tests in pytest
        assert pytest_plugin.in_pytest

    def test_in_pytest_pretend(self):
        # The test helper should work to ignore it
        with pretend_not_under_pytest():
            assert not pytest_plugin.in_pytest



from django.db import models


class Author(models.Model):
    name = models.CharField(max_length=32, unique=True)
    age = models.IntegerField()


class Book(models.Model):
    title = models.CharField(max_length=128)
    author = models.ForeignKey(Author, on_delete=models.CASCADE)


class Award(models.Model):
    name = models.CharField(max_length=128)
    author = models.ForeignKey(Author, on_delete=models.CASCADE)


class Contract(models.Model):
    amount = models.IntegerField()
    author = models.ManyToManyField(Author)



import errno
import os
import shutil
from contextlib import contextmanager

from django.db import connections

from django_perf_rec import pytest_plugin


def run_query(alias, sql, params=None):
    with connections[alias].cursor() as cursor:
        cursor.execute(sql, params)


@contextmanager
def temporary_path(path):
    ensure_path_does_not_exist(path)
    yield
    ensure_path_does_not_exist(path)


def ensure_path_does_not_exist(path):
    if path.endswith("/"):
        shutil.rmtree(path, ignore_errors=True)
    else:
        try:
            os.unlink(path)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise


@contextmanager
def pretend_not_under_pytest():
    orig = pytest_plugin.in_pytest
    pytest_plugin.in_pytest = False
    try:
        yield
    finally:
        pytest_plugin.in_pytest = orig



import errno
import os

import yaml
from django.core.files import locks


class KVFile:
    def __init__(self, file_name):
        self.file_name = file_name
        self.data = self.load(file_name)

    def __len__(self):
        return len(self.data)

    LOAD_CACHE = {}

    @classmethod
    def load(cls, file_name):
        if file_name not in cls.LOAD_CACHE:
            cls.LOAD_CACHE[file_name] = cls.load_file(file_name)
        return cls.LOAD_CACHE[file_name]

    @classmethod
    def load_file(cls, file_name):
        try:
            with open(file_name) as fp:
                locks.lock(fp, locks.LOCK_EX)
                content = fp.read()
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                content = "{}"
            else:
                raise

        data = yaml.safe_load(content)

        if data is None:
            return {}
        elif not isinstance(data, dict):
            raise TypeError("YAML content of {} is not a dictionary".format(file_name))

        return data

    @classmethod
    def _clear_load_cache(cls):
        # Should really only be used in testing this class
        cls.LOAD_CACHE = {}

    def get(self, key, default):
        return self.data.get(key, default)

    def set_and_save(self, key, value):
        if self.data.get(key, object()) == value:
            return

        fd = os.open(self.file_name, os.O_RDWR | os.O_CREAT, mode=0o666)
        with os.fdopen(fd, "r+") as fp:
            locks.lock(fd, locks.LOCK_EX)

            data = yaml.safe_load(fp)
            if data is None:
                data = {}

            self.data[key] = value
            data[key] = value

            fp.seek(0)
            yaml.safe_dump(
                data, fp, default_flow_style=False, allow_unicode=True, width=10000
            )



from functools import wraps
from types import MethodType

from django.conf import settings
from django.db import connections

from django_perf_rec.orm import patch_ORM_to_be_deterministic
from django_perf_rec.settings import perf_rec_settings
from django_perf_rec.sql import sql_fingerprint
from django_perf_rec.utils import sorted_names


class DBOp:
    def __init__(self, alias, sql):
        self.alias = alias
        self.sql = sql

    def __repr__(self):
        return "DBOp({!r}, {!r})".format(repr(self.alias), repr(self.sql))

    def __eq__(self, other):
        return (
            isinstance(other, DBOp)
            and self.alias == other.alias
            and self.sql == other.sql
        )


class DBRecorder:
    """
    Monkey-patch-wraps a database connection to call 'callback' on every
    query it runs.
    """

    def __init__(self, alias, callback):
        self.alias = alias
        self.callback = callback

    def __enter__(self):
        """
        When using the debug cursor wrapper, Django calls
        connection.ops.last_executed_query to get the SQL from the client
        library. Here we wrap this function on the connection to grab the SQL
        as it comes out.
        """
        patch_ORM_to_be_deterministic()

        connection = connections[self.alias]
        self.orig_force_debug_cursor = connection.force_debug_cursor
        connection.force_debug_cursor = True

        def call_callback(func):
            alias = self.alias
            callback = self.callback

            @wraps(func)
            def inner(self, *args, **kwargs):
                sql = func(*args, **kwargs)
                hide_columns = perf_rec_settings.HIDE_COLUMNS
                callback(
                    DBOp(
                        alias=alias, sql=sql_fingerprint(sql, hide_columns=hide_columns)
                    )
                )
                return sql

            return inner

        self.orig_last_executed_query = connection.ops.last_executed_query
        connection.ops.last_executed_query = MethodType(
            call_callback(connection.ops.last_executed_query), connection.ops
        )

    def __exit__(self, exc_type, exc_value, exc_traceback):
        connection = connections[self.alias]
        connection.force_debug_cursor = False
        connection.ops.last_executed_query = self.orig_last_executed_query


class AllDBRecorder:
    """
    Launches DBRecorders on all database connections
    """

    def __init__(self, callback):
        self.callback = callback

    def __enter__(self):
        self.recorders = []
        for alias in sorted_names(settings.DATABASES.keys()):
            recorder = DBRecorder(alias, self.callback)
            recorder.__enter__()
            self.recorders.append(recorder)

    def __exit__(self, type_, value, traceback):
        for recorder in reversed(self.recorders):
            recorder.__exit__(type_, value, traceback)
        self.recorders = []



import inspect
import re
from collections.abc import Mapping, Sequence
from functools import wraps
from types import MethodType

from django.conf import settings
from django.core.cache import caches

from django_perf_rec.utils import sorted_names


class CacheOp:
    def __init__(self, alias, operation, key_or_keys):
        self.alias = alias
        self.operation = operation
        if isinstance(key_or_keys, str):
            self.key_or_keys = self.clean_key(key_or_keys)
        elif isinstance(key_or_keys, (Mapping, Sequence)):
            self.key_or_keys = sorted(self.clean_key(k) for k in key_or_keys)
        else:
            raise ValueError("key_or_keys must be a string, mapping, or sequence")

    @classmethod
    def clean_key(cls, key):
        """
        Replace things that look like variables with a '#' so tests aren't
        affected by random variables
        """
        for var_re in cls.VARIABLE_RES:
            key = var_re.sub("#", key)
        return key

    VARIABLE_RES = (
        # Django session keys for 'cache' backend
        re.compile(r"(?<=django\.contrib\.sessions\.cache)[0-9a-z]{32}\b"),
        # Django session keys for 'cached_db' backend
        re.compile(r"(?<=django\.contrib\.sessions\.cached_db)[0-9a-z]{32}\b"),
        # Long random hashes
        re.compile(r"\b[0-9a-f]{32}\b"),
        # UUIDs
        re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"),
        # Integers
        re.compile(r"\d+"),
    )

    def __eq__(self, other):
        return (
            isinstance(other, CacheOp)
            and self.alias == other.alias
            and self.operation == other.operation
            and self.key_or_keys == other.key_or_keys
        )


class CacheRecorder:
    """
    Monkey patches a cache class to call 'callback' on every operation it calls
    """

    def __init__(self, alias, callback):
        self.alias = alias
        self.callback = callback

    def __enter__(self):
        cache = caches[self.alias]

        def call_callback(func):
            alias = self.alias
            callback = self.callback

            @wraps(func)
            def inner(self, *args, **kwargs):
                # Ignore operations from the cache class calling itself

                # Get the self of the parent via stack inspection
                frame = inspect.currentframe()
                try:
                    frame = frame.f_back
                    is_internal_call = frame.f_locals.get("self", None) is self
                finally:
                    # Always delete frame references to help garbage collector
                    del frame

                if not is_internal_call:
                    callback(
                        CacheOp(
                            alias=alias,
                            operation=str(func.__name__),
                            key_or_keys=args[0],
                        )
                    )

                return func(*args, **kwargs)

            return inner

        self.orig_methods = {name: getattr(cache, name) for name in self.cache_methods}
        for name in self.cache_methods:
            orig_method = self.orig_methods[name]
            setattr(cache, name, MethodType(call_callback(orig_method), cache))

    def __exit__(self, _, __, ___):
        cache = caches[self.alias]
        for name in self.cache_methods:
            setattr(cache, name, self.orig_methods[name])
        del self.orig_methods

    cache_methods = (
        "add",
        "decr",
        "delete",
        "delete_many",
        "get",
        "get_many",
        "incr",
        "set",
        "set_many",
    )


class AllCacheRecorder:
    """
    Launches CacheRecorders on all the active caches
    """

    def __init__(self, callback):
        self.callback = callback

    def __enter__(self):
        self.recorders = []
        for name in sorted_names(settings.CACHES.keys()):
            recorder = CacheRecorder(name, self.callback)
            recorder.__enter__()
            self.recorders.append(recorder)

    def __exit__(self, type_, value, traceback):
        for recorder in reversed(self.recorders):
            recorder.__exit__(type_, value, traceback)
        self.recorders = []



import shutil
from tempfile import mkdtemp

import pytest
import yaml
from django.test import SimpleTestCase

from django_perf_rec.yaml import KVFile


class KVFileTests(SimpleTestCase):
    def setUp(self):
        super().setUp()
        KVFile._clear_load_cache()
        self.temp_dir = mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        super().tearDown()

    def test_load_no_permissions(self):
        with pytest.raises(IOError):
            KVFile("/")

    def test_load_non_existent_is_empty(self):
        kvf = KVFile(self.temp_dir + "/foo.yml")
        assert len(kvf) == 0
        default = object()
        assert kvf.get("foo", default) is default

    def test_load_existent(self):
        file_name = self.temp_dir + "/foo.yml"
        with open(file_name, "w") as fp:
            fp.write("foo: bar")

        kvf = KVFile(file_name)
        assert len(kvf) == 1
        assert kvf.get("foo", "") == "bar"

    def test_load_empty(self):
        file_name = self.temp_dir + "/foo.yml"
        with open(file_name, "w") as fp:
            fp.write("")

        assert len(KVFile(file_name)) == 0

    def test_load_whitespace_empty(self):
        file_name = self.temp_dir + "/foo.yml"
        with open(file_name, "w") as fp:
            fp.write(" \n")

        assert len(KVFile(file_name)) == 0

    def test_load_non_dictionary(self):
        file_name = self.temp_dir + "/foo.yml"
        with open(file_name, "w") as fp:
            fp.write("[not, a, dictionary]")

        with pytest.raises(TypeError) as excinfo:
            KVFile(file_name)
        assert "not a dictionary" in str(excinfo.value)

    def test_get_after_set_same(self):
        kvf = KVFile(self.temp_dir + "/foo.yml")
        kvf.set_and_save("foo", "bar")

        assert len(kvf) == 1
        assert kvf.get("foo", "") == "bar"

    def test_load_second_same(self):
        kvf = KVFile(self.temp_dir + "/foo.yml")
        kvf.set_and_save("foo", "bar")
        kvf2 = KVFile(self.temp_dir + "/foo.yml")

        assert len(kvf2) == 1
        assert kvf2.get("foo", "") == "bar"

    def test_sets_dont_cause_append_duplication(self):
        file_name = self.temp_dir + "/foo.yml"
        kvf = KVFile(file_name)
        kvf.set_and_save("foo", "bar")
        kvf.set_and_save("foo2", "bar")

        with open(file_name) as fp:
            lines = fp.readlines()
            fp.seek(0)
            data = yaml.safe_load(fp)

        assert len(lines) == 2
        assert data == {"foo": "bar", "foo2": "bar"}



import django
import patchy
from django.db.models.query import QuerySet
from django.db.models.query_utils import Q
from django.db.models.sql.query import Query


def patch_ORM_to_be_deterministic():
    """
    Django's ORM is non-deterministic with regards to the queries it outputs
    for e.g. OR clauses. We need it to be deterministic so that we can compare
    queries between runs, so we make a couple patches to its internals to do
    this. Mostly this is done by adding sorted() in some places so we're not
    affected by the vagaries of random dict iteration order.

    There is no undo for this, but it doesn't make the ORM much slower or
    anything bad.
    """
    if patch_ORM_to_be_deterministic.have_patched:
        return
    patch_ORM_to_be_deterministic.have_patched = True

    patch_QuerySet()
    patch_Query()
    patch_Q()


patch_ORM_to_be_deterministic.have_patched = False


def patch_QuerySet():
    patchy.patch(
        QuerySet.annotate,
        """\
        @@ -17,7 +17,7 @@
                 except (AttributeError, TypeError):
                     raise TypeError("Complex annotations require an alias")
                 annotations[arg.default_alias] = arg
        -    annotations.update(kwargs)
        +    annotations.update(sorted(kwargs.items()))

             clone = self._clone()
             names = self._fields
    """,
    )


if django.VERSION >= (2, 2):

    def patch_Query():
        patchy.patch(
            Query.add_extra,
            """\
            @@ -13,7 +13,7 @@
                         param_iter = iter(select_params)
                     else:
                         param_iter = iter([])
            -        for name, entry in select.items():
            +        for name, entry in sorted(select.items()):
                         entry = str(entry)
                         entry_params = []
                         pos = entry.find("%s")
        """,
        )


else:

    def patch_Query():
        patchy.patch(
            Query.add_extra,
            """\
            @@ -13,7 +13,7 @@
                         param_iter = iter(select_params)
                     else:
                         param_iter = iter([])
            -        for name, entry in select.items():
            +        for name, entry in sorted(select.items()):
                         entry = force_text(entry)
                         entry_params = []
                         pos = entry.find("%s")
        """,
        )


if django.VERSION < (2, 0, 3):

    def patch_Q():
        # This one can't be done by patchy since __init__ is different in Python 3,
        # maybe one day https://github.com/adamchainz/patchy/issues/31 will be
        # fixed.
        def __init__(self, *args, **kwargs):
            connector = kwargs.pop("_connector", None)
            negated = kwargs.pop("_negated", False)
            super().__init__(
                children=list(args) + sorted(kwargs.items()),
                connector=connector,
                negated=negated,
            )

        Q.__init__ = __init__


else:

    # After Django 2.0.3, kwargs are sorted so we don't need to patch anything
    # https://code.djangoproject.com/ticket/29125
    def patch_Q():
        pass



import pytest

from django_perf_rec import get_perf_path, get_record_name, record
from tests.utils import run_query

pytestmark = [pytest.mark.django_db]


@pytest.fixture
def record_auto_name():
    with record():
        yield


def test_auto_name(record_auto_name):
    run_query("default", "SELECT 1337")


@pytest.fixture
def record_auto_name_with_request(request):
    with record():
        yield


def test_auto_name_with_request(record_auto_name_with_request):
    run_query("default", "SELECT 1337")


@pytest.fixture
def record_build_name(request):
    record_name = get_record_name(
        class_name=request.cls.__name__ if request.cls is not None else None,
        test_name=request.function.__name__,
    )
    path = get_perf_path(file_path=request.fspath.strpath)
    with record(record_name=record_name, path=path):
        yield


def test_build_name(record_build_name):
    run_query("default", "SELECT 1337")



from django.test import SimpleTestCase

from django_perf_rec.utils import current_test, sorted_names


class CurrentTestTests(SimpleTestCase):
    def test_here(self):
        details = current_test()
        assert details.file_path == __file__
        assert details.class_name == "CurrentTestTests"
        assert details.test_name == "test_here"

    def test_twice_same(self):
        assert current_test() == current_test()

    def test_functional(self):
        def test_thats_functional():
            return current_test()

        details = test_thats_functional()
        assert details.file_path == __file__
        assert details.class_name is None
        assert details.test_name == "test_thats_functional"


class SortedNamesTests(SimpleTestCase):
    def test_empty(self):
        assert sorted_names([]) == []

    def test_just_default(self):
        assert sorted_names(["default"]) == ["default"]

    def test_just_something(self):
        assert sorted_names(["something"]) == ["something"]

    def test_does_sort(self):
        assert sorted_names(["b", "a"]) == ["a", "b"]

    def test_sort_keeps_default_first(self):
        assert sorted_names(["a", "default"]) == ["default", "a"]



from django_perf_rec.sql import sql_fingerprint


def test_empty():
    assert sql_fingerprint("") == ""
    assert sql_fingerprint("\n\n    \n") == ""


def test_select():
    assert sql_fingerprint("SELECT `f1`, `f2` FROM `b`") == "SELECT ... FROM `b`"


def test_select_show_columns(settings):
    assert (
        sql_fingerprint("SELECT `f1`, `f2` FROM `b`", hide_columns=False)
        == "SELECT `f1`, `f2` FROM `b`"
    )


def test_select_limit(settings):
    assert (
        sql_fingerprint("SELECT `f1`, `f2` FROM `b` LIMIT 12", hide_columns=False)
        == "SELECT `f1`, `f2` FROM `b` LIMIT #"
    )


def test_select_coalesce_show_columns(settings):
    assert (
        sql_fingerprint(
            (
                "SELECT `table`.`f1`, COALESCE(table.f2->>'a', table.f2->>'b', "
                + "'default') FROM `table`"
            ),
            hide_columns=False,
        )
        == "SELECT `table`.`f1`, COALESCE(table.f2->>#, table.f2->>#, #) FROM `table`"
    )


def test_select_where():
    assert (
        sql_fingerprint(
            "SELECT DISTINCT `table`.`field` FROM `table` WHERE `table`.`id` = 1"
        )
        == "SELECT DISTINCT `table`.`field` FROM `table` WHERE `table`.`id` = #"
    )


def test_select_where_show_columns(settings):
    assert (
        sql_fingerprint(
            "SELECT DISTINCT `table`.`field` FROM `table` WHERE `table`.`id` = 1",
            hide_columns=False,
        )
        == "SELECT DISTINCT `table`.`field` FROM `table` WHERE `table`.`id` = #"
    )


def test_select_comment():
    assert (
        sql_fingerprint("SELECT /* comment */ `f1`, `f2` FROM `b`")
        == "SELECT /* comment */ ... FROM `b`"
    )


def test_select_comment_show_columns(settings):
    assert (
        sql_fingerprint("SELECT /* comment */ `f1`, `f2` FROM `b`", hide_columns=False)
        == "SELECT /* comment */ `f1`, `f2` FROM `b`"
    )


def test_select_join():
    assert (
        sql_fingerprint(
            "SELECT f1, f2 FROM a INNER JOIN b ON (a.b_id = b.id) WHERE a.f2 = 1"
        )
        == "SELECT ... FROM a INNER JOIN b ON (a.b_id = b.id) WHERE a.f2 = #"
    )


def test_select_join_show_columns(settings):
    assert (
        sql_fingerprint(
            "SELECT f1, f2 FROM a INNER JOIN b ON (a.b_id = b.id) WHERE a.f2 = 1",
            hide_columns=False,
        )
        == "SELECT f1, f2 FROM a INNER JOIN b ON (a.b_id = b.id) WHERE a.f2 = #"
    )


def test_select_order_by():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a ORDER BY f3")
        == "SELECT ... FROM a ORDER BY f3"
    )


def test_select_order_by_limit():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a ORDER BY f3 LIMIT 12")
        == "SELECT ... FROM a ORDER BY f3 LIMIT #"
    )


def test_select_order_by_show_columns(settings):
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a ORDER BY f3", hide_columns=False)
        == "SELECT f1, f2 FROM a ORDER BY f3"
    )


def test_select_order_by_multiple():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a ORDER BY f3, f4")
        == "SELECT ... FROM a ORDER BY f3, f4"
    )


def test_select_group_by():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a GROUP BY f1")
        == "SELECT ... FROM a GROUP BY f1"
    )


def test_select_group_by_show_columns(settings):
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a GROUP BY f1", hide_columns=False)
        == "SELECT f1, f2 FROM a GROUP BY f1"
    )


def test_select_group_by_multiple():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a GROUP BY f1, f2")
        == "SELECT ... FROM a GROUP BY f1, f2"
    )


def test_select_group_by_having():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a GROUP BY f1 HAVING f1 > 21")
        == "SELECT ... FROM a GROUP BY f1 HAVING f1 > #"
    )


def test_select_group_by_having_show_columns(settings):
    assert (
        sql_fingerprint(
            "SELECT f1, f2 FROM a GROUP BY f1 HAVING f1 > 21", hide_columns=False
        )
        == "SELECT f1, f2 FROM a GROUP BY f1 HAVING f1 > #"
    )


def test_select_group_by_having_multiple():
    assert (
        sql_fingerprint("SELECT f1, f2 FROM a GROUP BY f1 HAVING f1 > 21, f2 < 42")
        == "SELECT ... FROM a GROUP BY f1 HAVING f1 > #, f2 < #"
    )


def test_insert():
    assert (
        sql_fingerprint("INSERT INTO `table` (`f1`, `f2`) VALUES ('v1', 2)")
        == "INSERT INTO `table` (...) VALUES (...)"
    )


def test_insert_show_columns(settings):
    assert (
        sql_fingerprint(
            "INSERT INTO `table` (`f1`, `f2`) VALUES ('v1', 2)", hide_columns=False
        )
        == "INSERT INTO `table` (`f1`, `f2`) VALUES (#, #)"
    )


def test_update():
    assert (
        sql_fingerprint("UPDATE `table` SET `foo` = 'bar' WHERE `table`.`id` = 1")
        == "UPDATE `table` SET ... WHERE `table`.`id` = #"
    )


def test_declare_cursor():
    assert (
        sql_fingerprint(
            'DECLARE "_django_curs_140239496394496_1300" NO SCROLL CURSOR WITHOUT'
        )
        == 'DECLARE "_django_curs_#" NO SCROLL CURSOR WITHOUT'
    )


def test_savepoint():
    assert sql_fingerprint("SAVEPOINT `s140323809662784_x54`") == "SAVEPOINT `#`"


def test_rollback_to_savepoint():
    assert (
        sql_fingerprint("ROLLBACK TO SAVEPOINT `s140323809662784_x54`")
        == "ROLLBACK TO SAVEPOINT `#`"
    )


def test_release_savepoint():
    assert (
        sql_fingerprint("RELEASE SAVEPOINT `s140699855320896_x17`")
        == "RELEASE SAVEPOINT `#`"
    )


def test_null_value():
    assert (
        sql_fingerprint(
            "SELECT `f1`, `f2` FROM `b` WHERE `b`.`name` IS NULL", hide_columns=False
        )
        == "SELECT `f1`, `f2` FROM `b` WHERE `b`.`name` IS #"
    )


def test_strip_duplicate_whitespaces():
    assert (
        sql_fingerprint(
            "SELECT    `f1`,  `f2` FROM  `b` WHERE   `b`.`f1` IS  NULL LIMIT 12  "
        )
        == "SELECT ... FROM `b` WHERE `b`.`f1` IS # LIMIT #"
    )


def test_strip_duplicate_whitespaces_recursive():
    assert (
        sql_fingerprint(
            "SELECT    `f1`,  `f2`, (   COALESCE(b.f3->>'en',   b.f3->>'fr', '')) "
            "FROM  `b` WHERE   (`b`.`f1` IS   NULL OR (  EXISTS COUNT(1) )) LIMIT 12  ",
            hide_columns=False,
        )
        == "SELECT `f1`, `f2`, (COALESCE(b.f3->>#, b.f3->>#, #)) "
        "FROM `b` WHERE (`b`.`f1` IS # OR (EXISTS COUNT(#))) LIMIT #"
    )


def test_strip_newlines():
    assert (
        sql_fingerprint("SELECT `f1`, `f2`\nFROM `b`\n LIMIT 12\n\n")
        == "SELECT ... FROM `b` LIMIT #"
    )


def test_strip_raw_query():
    assert (
        sql_fingerprint(
            """
SELECT 'f1'
    , 'f2'
    , 'f3'
FROM "table_a" WHERE "table_a"."f1" = 1 OR (
"table_a"."type" = 'A' AND
EXISTS (
    SELECT "table_b"."id"
    FROM "table_b"
    WHERE "table_b"."id" = 1
) = true)
"""
        )
        == (
            'SELECT ... FROM "table_a" WHERE "table_a"."f1" = # OR '
            + '("table_a"."type" = # AND EXISTS (SELECT "table_b"."id" FROM '
            + '"table_b" WHERE "table_b"."id" = # ) = true)'
        )
    )


def test_in_single_value():
    assert (
        sql_fingerprint("SELECT `f1`, `f2` FROM `b` WHERE `x` IN (1)")
        == "SELECT ... FROM `b` WHERE `x` IN (...)"
    )


def test_in_multiple_values():
    assert (
        sql_fingerprint("SELECT `f1`, `f2` FROM `b` WHERE `x` IN (1, 2, 3)")
        == "SELECT ... FROM `b` WHERE `x` IN (...)"
    )


def test_in_multiple_clauses():
    assert (
        sql_fingerprint(
            "SELECT `f1`, `f2` FROM `b` WHERE `x` IN (1, 2, 3) AND `y` IN (4, 5, 6)"
        )
        == "SELECT ... FROM `b` WHERE `x` IN (...) AND `y` IN (...)"
    )


def test_in_multiple_values_and_clause():
    assert (
        sql_fingerprint(
            "SELECT `f1`, `f2` FROM `b` WHERE `x` IN (1, 2, 3) AND (`y` = 1 OR `y` = 2)"
        )
        == "SELECT ... FROM `b` WHERE `x` IN (...) AND (`y` = # OR `y` = #)"
    )


def test_in_subquery():
    assert (
        sql_fingerprint("SELECT `f1`, `f2` FROM `b` WHERE `x` IN (SELECT 1)")
        == "SELECT ... FROM `b` WHERE `x` IN (SELECT #)"
    )



import pytest
from django.core.cache import caches
from django.test import SimpleTestCase, TestCase

from django_perf_rec.cache import AllCacheRecorder, CacheOp, CacheRecorder

try:
    from unittest import mock
except ImportError:
    import mock


class CacheOpTests(SimpleTestCase):
    def test_clean_key_integer(self):
        assert CacheOp.clean_key("foo1") == "foo#"

    def test_clean_key_uuid(self):
        assert CacheOp.clean_key("bdfc9986-d461-4a5e-bf98-8688993abcfb") == "#"

    def test_clean_key_random_hash(self):
        assert CacheOp.clean_key("abc123abc123abc123abc123abc12345") == "#"

    def test_clean_key_session_key_cache_backend(self):
        key = "django.contrib.sessions.cacheabcdefghijklmnopqrstuvwxyz012345"
        assert CacheOp.clean_key(key) == "django.contrib.sessions.cache#"

    def test_clean_key_session_key_cached_db_backend(self):
        key = "django.contrib.sessions.cached_db" + "abcdefghijklmnopqrstuvwxyz012345"
        assert CacheOp.clean_key(key) == "django.contrib.sessions.cached_db#"

    def test_key(self):
        op = CacheOp("default", "foo", "bar")
        assert op.alias == "default"
        assert op.operation == "foo"
        assert op.key_or_keys == "bar"

    def test_keys(self):
        op = CacheOp("default", "foo", ["bar", "baz"])
        assert op.alias == "default"
        assert op.operation == "foo"
        assert op.key_or_keys == ["bar", "baz"]

    def test_invalid(self):
        with pytest.raises(ValueError):
            CacheOp("x", "foo", object())

    def test_equal(self):
        assert CacheOp("x", "foo", "bar") == CacheOp("x", "foo", "bar")

    def test_not_equal_alias(self):
        assert CacheOp("x", "foo", "bar") != CacheOp("y", "foo", "bar")

    def test_not_equal_operation(self):
        assert CacheOp("x", "foo", "bar") != CacheOp("x", "bar", "bar")

    def test_not_equal_keys(self):
        assert CacheOp("x", "foo", ["bar"]) != CacheOp("x", "foo", ["baz"])


class CacheRecorderTests(TestCase):
    def test_default(self):
        callback = mock.Mock()
        with CacheRecorder("default", callback):
            caches["default"].get("foo")
        callback.assert_called_once_with(CacheOp("default", "get", "foo"))

    def test_secondary(self):
        callback = mock.Mock()
        with CacheRecorder("second", callback):
            caches["second"].get("foo")
        callback.assert_called_once_with(CacheOp("second", "get", "foo"))

    def test_secondary_default_not_recorded(self):
        callback = mock.Mock()
        with CacheRecorder("second", callback):
            caches["default"].get("foo")
        assert len(callback.mock_calls) == 0


class AllCacheRecorderTests(TestCase):
    def test_records_all(self):
        callback = mock.Mock()
        with AllCacheRecorder(callback):
            caches["default"].get("foo")
            caches["second"].set("bar", "baz")
            caches["default"].delete_many(["foo"])

        assert callback.mock_calls == [
            mock.call(CacheOp("default", "get", "foo")),
            mock.call(CacheOp("second", "set", "bar")),
            mock.call(CacheOp("default", "delete_many", ["foo"])),
        ]



from django.db.models import Q
from django.test import SimpleTestCase

from django_perf_rec.orm import patch_ORM_to_be_deterministic


class PatchORMToBeDeterministicTests(SimpleTestCase):
    def test_call_it(self):
        patch_ORM_to_be_deterministic()

    def test_call_it_again(self):
        patch_ORM_to_be_deterministic()

    def test_q_connector(self):
        q1 = Q(foo="bar") | Q(bar="foo")
        _path, args, kwargs = q1.deconstruct()
        q2 = Q(*args, **kwargs)
        self.assertEqual(q1, q2)



import os

import pytest
from django.core.cache import caches
from django.db.models import Q
from django.db.models.functions import Upper
from django.test import SimpleTestCase, TestCase, override_settings

from django_perf_rec import TestCaseMixin, get_perf_path, get_record_name, record
from tests.testapp.models import Author
from tests.utils import pretend_not_under_pytest, run_query, temporary_path

FILE_DIR = os.path.dirname(__file__)


class RecordTests(TestCase):
    def test_single_db_query(self):
        with record():
            run_query("default", "SELECT 1337")

    def test_single_db_query_model(self):
        with record():
            list(Author.objects.all())

    @override_settings(PERF_REC={"HIDE_COLUMNS": False})
    def test_single_db_query_model_with_columns(self):
        with record():
            list(Author.objects.all())

    def test_multiple_db_queries(self):
        with record():
            run_query("default", "SELECT 1337")
            run_query("default", "SELECT 4949")

    def test_non_deterministic_QuerySet_annotate(self):
        with record():
            list(Author.objects.annotate(x=Upper("name"), y=Upper("name")))

    def test_non_deterministic_QuerySet_extra(self):
        with record():
            list(Author.objects.extra(select={"x": "1", "y": "1"}))

    def test_non_deterministic_Q_query(self):
        with record():
            list(Author.objects.filter(Q(name="foo", age=1)))

    def test_single_cache_op(self):
        with record():
            caches["default"].get("foo")

    def test_multiple_cache_ops(self):
        with record():
            caches["default"].set("foo", "bar")
            caches["second"].get_many(["foo", "bar"])
            caches["default"].delete("foo")

    def test_multiple_calls_in_same_function_are_different_records(self):
        with record():
            caches["default"].get("foo")

        with record():
            caches["default"].get("bar")

    def test_custom_name(self):
        with record(record_name="custom"):
            caches["default"].get("foo")

    def test_custom_name_multiple_calls(self):
        with record(record_name="custom"):
            caches["default"].get("foo")

        with pytest.raises(AssertionError) as excinfo:
            with record(record_name="custom"):
                caches["default"].get("bar")

        assert "Performance record did not match" in str(excinfo.value)

    def test_diff(self):
        with pretend_not_under_pytest():
            with record(record_name="test_diff"):
                caches["default"].get("foo")

            with pytest.raises(AssertionError) as excinfo:
                with record(record_name="test_diff"):
                    caches["default"].get("bar")

            msg = str(excinfo.value)
            assert "- cache|get: foo\n" in msg
            assert "+ cache|get: bar\n" in msg

    def test_path_pointing_to_filename(self):
        with temporary_path("custom.perf.yml"):

            with record(path="custom.perf.yml"):
                caches["default"].get("foo")

            assert os.path.exists("custom.perf.yml")

    def test_path_pointing_to_filename_record_twice(self):
        with temporary_path("custom.perf.yml"):

            with record(path="custom.perf.yml"):
                caches["default"].get("foo")

            with record(path="custom.perf.yml"):
                caches["default"].get("foo")

    def test_path_pointing_to_dir(self):
        temp_dir = os.path.join(FILE_DIR, "perf_files/")
        with temporary_path(temp_dir):

            with record(path="perf_files/"):
                caches["default"].get("foo")

            full_path = os.path.join(FILE_DIR, "perf_files", "test_api.perf.yml")
            assert os.path.exists(full_path)

    def test_custom_nested_path(self):
        temp_dir = os.path.join(FILE_DIR, "perf_files/")
        with temporary_path(temp_dir):

            with record(path="perf_files/api/"):
                caches["default"].get("foo")

            full_path = os.path.join(FILE_DIR, "perf_files", "api", "test_api.perf.yml")
            assert os.path.exists(full_path)

    @override_settings(PERF_REC={"MODE": "once"})
    def test_mode_once(self):
        temp_dir = os.path.join(FILE_DIR, "perf_files/")
        with temporary_path(temp_dir):

            with record(path="perf_files/api/"):
                caches["default"].get("foo")

            full_path = os.path.join(FILE_DIR, "perf_files", "api", "test_api.perf.yml")
            assert os.path.exists(full_path)

    @override_settings(PERF_REC={"MODE": "none"})
    def test_mode_none(self):
        temp_dir = os.path.join(FILE_DIR, "perf_files/")
        with temporary_path(temp_dir):

            with pytest.raises(AssertionError) as excinfo:

                with record(path="perf_files/api/"):
                    caches["default"].get("foo")

            assert "Original performance record does not exist" in str(excinfo.value)

            full_path = os.path.join(FILE_DIR, "perf_files", "api", "test_api.perf.yml")
            assert not os.path.exists(full_path)

    @override_settings(PERF_REC={"MODE": "all"})
    def test_mode_all(self):
        temp_dir = os.path.join(FILE_DIR, "perf_files/")
        with temporary_path(temp_dir):

            with pytest.raises(AssertionError) as excinfo:

                with record(path="perf_files/api/"):
                    caches["default"].get("foo")

            assert "Original performance record did not exist" in str(excinfo.value)

            full_path = os.path.join(FILE_DIR, "perf_files", "api", "test_api.perf.yml")
            assert os.path.exists(full_path)

    def test_delete_on_cascade_called_twice(self):
        arthur = Author.objects.create(name="Arthur", age=42)
        with record():
            arthur.delete()


class GetPerfPathTests(SimpleTestCase):
    def test_py_file(self):
        assert get_perf_path("foo.py") == "foo.perf.yml"

    def test_pyc_file(self):
        assert get_perf_path("foo.pyc") == "foo.perf.yml"

    def test_unknown_file(self):
        assert get_perf_path("foo.plob") == "foo.plob.perf.yml"


class GetRecordNameTests(SimpleTestCase):
    def test_class_and_test(self):
        assert (
            get_record_name(class_name="FooTests", test_name="test_bar")
            == "FooTests.test_bar"
        )

    def test_just_test(self):
        assert get_record_name(test_name="test_baz") == "test_baz"

    def test_multiple_calls(self):
        assert get_record_name(test_name="test_qux") == "test_qux"
        assert get_record_name(test_name="test_qux") == "test_qux.2"

    def test_multiple_calls_from_different_files(self):
        assert get_record_name(test_name="test_qux", file_name="foo.py") == "test_qux"

        assert get_record_name(test_name="test_qux", file_name="foo2.py") == "test_qux"

        assert get_record_name(test_name="test_qux", file_name="foo.py") == "test_qux"

        assert get_record_name(test_name="test_qux", file_name="foo.py") == "test_qux.2"


class TestCaseMixinTests(TestCaseMixin, TestCase):
    def test_record_performance(self):
        with self.record_performance():
            caches["default"].get("foo")

    def test_record_performance_record_name(self):
        with self.record_performance(record_name="other"):
            caches["default"].get("foo")

    def test_record_performance_file_name(self):
        perf_name = __file__.replace(".py", ".file_name.perf.yml")
        with self.record_performance(path=perf_name):
            caches["default"].get("foo")



import os
from threading import local

from django.core.cache import DEFAULT_CACHE_ALIAS
from django.db import DEFAULT_DB_ALIAS
from django.utils.functional import SimpleLazyObject

from django_perf_rec import pytest_plugin
from django_perf_rec.cache import AllCacheRecorder
from django_perf_rec.db import AllDBRecorder
from django_perf_rec.settings import perf_rec_settings
from django_perf_rec.utils import current_test, record_diff
from django_perf_rec.yaml import KVFile

record_current = local()


def record(*, record_name=None, path=None):
    # Lazy since we may not need this to determine record_name or path,
    # depending on logic below
    test_details = SimpleLazyObject(current_test)

    if path is None or path.endswith("/"):
        file_name = get_perf_path(test_details.file_path)
    else:
        file_name = path

    if path is not None and path.endswith("/"):
        if not os.path.isabs(path):
            directory = os.path.join(os.path.dirname(test_details.file_path), path)
            if not os.path.exists(directory):
                os.makedirs(directory)
        else:
            directory = path

        file_name = os.path.join(directory, os.path.basename(file_name))

    if record_name is None:
        record_name = get_record_name(
            test_name=test_details.test_name,
            class_name=test_details.class_name,
            file_name=file_name,
        )

    return PerformanceRecorder(file_name, record_name)


def get_perf_path(file_path):
    if file_path.endswith(".py"):
        perf_path = file_path[: -len(".py")] + ".perf.yml"
    elif file_path.endswith(".pyc"):
        perf_path = file_path[: -len(".pyc")] + ".perf.yml"
    else:
        perf_path = file_path + ".perf.yml"
    return perf_path


def get_record_name(test_name, class_name=None, file_name=""):
    if class_name:
        record_name = "{class_}.{test}".format(class_=class_name, test=test_name)
    else:
        record_name = test_name

    # Multiple calls inside the same test should end up suffixing with .2, .3 etc.
    record_spec = (file_name, record_name)
    if getattr(record_current, "record_spec", None) == record_spec:
        record_current.counter += 1
        record_name = record_name + ".{}".format(record_current.counter)
    else:
        record_current.record_spec = record_spec
        record_current.counter = 1

    return record_name


class PerformanceRecorder:
    def __init__(self, file_name, record_name):
        self.file_name = file_name
        self.record_name = record_name

        self.record = []
        self.db_recorder = AllDBRecorder(self.on_db_op)
        self.cache_recorder = AllCacheRecorder(self.on_cache_op)

    def __enter__(self):
        self.db_recorder.__enter__()
        self.cache_recorder.__enter__()
        self.load_recordings()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cache_recorder.__exit__(exc_type, exc_value, exc_traceback)
        self.db_recorder.__exit__(exc_type, exc_value, exc_traceback)

        if exc_type is None:
            self.save_or_assert()

    def on_db_op(self, db_op):
        name_parts = ["db"]
        if db_op.alias != DEFAULT_DB_ALIAS:
            name_parts.append(db_op.alias)
        name = "|".join(name_parts)

        self.record.append({name: db_op.sql})

    def on_cache_op(self, cache_op):
        name_parts = ["cache"]
        if cache_op.alias != DEFAULT_CACHE_ALIAS:
            name_parts.append(cache_op.alias)
        name_parts.append(cache_op.operation)
        name = "|".join(name_parts)

        self.record.append({name: cache_op.key_or_keys})

    def load_recordings(self):
        self.records_file = KVFile(self.file_name)

    def save_or_assert(self):
        orig_record = self.records_file.get(self.record_name, None)

        if perf_rec_settings.MODE == "none":
            assert (
                orig_record is not None
            ), "Original performance record does not exist for {}".format(
                self.record_name
            )

        if orig_record is not None:
            msg = "Performance record did not match for {}".format(self.record_name)
            if not pytest_plugin.in_pytest:
                msg += "\n{}".format(record_diff(orig_record, self.record))
            assert self.record == orig_record, msg

        self.records_file.set_and_save(self.record_name, self.record)

        if perf_rec_settings.MODE == "all":
            assert (
                orig_record is not None
            ), "Original performance record did not exist for {}".format(
                self.record_name
            )


class TestCaseMixin:
    """
    Adds record_performance() method to TestCase class it's mixed into
    for easy import-free use.
    """

    def record_performance(self, *, record_name=None, path=None):
        return record(record_name=record_name, path=path)



import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

DEBUG = True
TEMPLATE_DEBUG = DEBUG

SECRET_KEY = "NOTASECRET"

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
    "replica": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
        "TEST": {"MIRROR": "default"},
    },
    "second": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
}

CACHES = {
    "default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"},
    "second": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"},
}

ALLOWED_HOSTS = []

INSTALLED_APPS = ["django.contrib.auth", "django.contrib.contenttypes", "tests.testapp"]

MIDDLEWARE_CLASSES = (
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

ROOT_URLCONF = "tests.urls"
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]



import pytest

from django_perf_rec import record
from tests.utils import run_query

pytestmark = [pytest.mark.django_db]


def test_duplicate_name():
    with record():
        run_query("default", "SELECT 1337")



import pytest
from django.test.utils import override_settings

from django_perf_rec import record
from tests.utils import run_query

pytestmark = [pytest.mark.django_db]


@override_settings(PERF_REC={"MODE": "none"})
def test_duplicate_name():
    with record():
        run_query("default", "SELECT 1337")
        run_query("default", "SELECT 4997")
        run_query("default", "SELECT 4998")



from django.test import SimpleTestCase, TestCase

from django_perf_rec.db import AllDBRecorder, DBOp, DBRecorder
from tests.utils import run_query

try:
    from unittest import mock
except ImportError:
    import mock


class DBOpTests(SimpleTestCase):
    def test_create(self):
        op = DBOp("myalias", "SELECT 1")
        assert op.alias == "myalias"
        assert op.sql == "SELECT 1"

    def test_equal(self):
        assert DBOp("foo", "bar") == DBOp("foo", "bar")

    def test_not_equal_alias(self):
        assert DBOp("foo", "bar") != DBOp("baz", "bar")

    def test_not_equal_sql(self):
        assert DBOp("foo", "bar") != DBOp("foo", "baz")


class DBRecorderTests(TestCase):
    databases = ("default", "second", "replica")

    def test_default(self):
        callback = mock.Mock()
        with DBRecorder("default", callback):
            run_query("default", "SELECT 1")
        callback.assert_called_once_with(DBOp("default", "SELECT #"))

    def test_secondary(self):
        callback = mock.Mock()
        with DBRecorder("second", callback):
            run_query("second", "SELECT 1")
        callback.assert_called_once_with(DBOp("second", "SELECT #"))

    def test_replica(self):
        callback = mock.Mock()
        with DBRecorder("replica", callback):
            run_query("replica", "SELECT 1")
        callback.assert_called_once_with(DBOp("replica", "SELECT #"))

    def test_secondary_default_not_recorded(self):
        callback = mock.Mock()
        with DBRecorder("second", callback):
            run_query("default", "SELECT 1")
        assert len(callback.mock_calls) == 0


class AllDBRecorderTests(TestCase):
    databases = ("default", "second", "replica")

    def test_records_all(self):
        callback = mock.Mock()
        with AllDBRecorder(callback):
            run_query("replica", "SELECT 1")
            run_query("default", "SELECT 2")
            run_query("second", "SELECT 3")

        assert callback.mock_calls == [
            mock.call(DBOp("replica", "SELECT #")),
            mock.call(DBOp("default", "SELECT #")),
            mock.call(DBOp("second", "SELECT #")),
        ]



from django.conf import settings


class Settings:

    defaults = {"HIDE_COLUMNS": True, "MODE": "once"}

    def get_setting(self, key):
        try:
            return settings.PERF_REC[key]
        except (AttributeError, KeyError):
            return self.defaults.get(key, None)

    @property
    def HIDE_COLUMNS(self):
        return self.get_setting("HIDE_COLUMNS")

    @property
    def MODE(self):
        return self.get_setting("MODE")


perf_rec_settings = Settings()
